@@ -1,26 +1,32 @@
 <?php
 /**
  * Change LDAP Password plugin functions
- * @copyright &copy; 2000-2003 Simon Annetts <simon@ateb.co.uk>
- * @copyright &copy; 2006-2007 The SquirrelMail Project Team
+ * @copyright (c) 2000-2003 Simon Annetts <simon@ateb.co.uk>
+ * @copyright (c) 2006-2007 The SquirrelMail Project Team
+ * @copyright (c) 2007 The NaSMail Project
  * @license http://opensource.org/licenses/gpl-license.php GNU Public License
- * @version $Id: functions.php,v 1.7 2007/01/20 11:51:22 tokul Exp $
+ * @version $Id: functions.php,v 1.12 2007/07/21 06:56:42 tokul Exp $
  * @package plugins
  * @subpackage change_ldappass
  */
 
+/** load configuration */
+include_once(SM_PATH . 'plugins/change_ldappass/load_config.php');
+
 /**
  * Checks user input
  * @return array
  */
 function change_ldappass_check() {
+    global $lcp_crack_dict;
+
     $cp_oldpass = '';
     $cp_newpass = '';
     $cp_verify = '';
     sqgetGlobalVar('cp_oldpass',$cp_oldpass,SQ_POST);
     sqgetGlobalVar('cp_newpass',$cp_newpass,SQ_POST);
     sqgetGlobalVar('cp_verify',$cp_verify,SQ_POST);
- 
+
     $Messages = array();
     if (! extension_loaded('ldap')) {
         array_push($Messages,_("PHP LDAP extension is not available."));
@@ -40,13 +46,32 @@ function change_ldappass_check() {
         array_push($Messages, _("Your new password doesn't match the verify password."));
     }
 
-    if (!preg_match("/^[A-Za-z0-9\%\^\*\(\)\-\_\+\=\[\]\{\}\:\@\#\~\,\.\?]+$/",$cp_newpass)) {
+    if (!preg_match("/^[A-Za-z0-9_%=:@#~,\\^\\*\\(\\)\\-\\+\\[\\]\\{\\}\\.\\?]+$/",$cp_newpass)) {
         // i18n: comma separated list of acceptable characters is listed in next line
-        array_push($Messages, 
+        array_push($Messages,
                    _("Passwords can only contain the following characters:"),
                    "A-Z, a-z, 0-9, %^*()-_+=[]{}:@#~,.?");
     }
 
+    // PHP crack support from NaSMail change_password plugin
+    if (! empty($lcp_crack_dict)) {
+        if (! function_exists('crack_opendict')) {
+            $Messages[] = _("PHP Crack extension is not available.");
+        } else {
+            // if crack_opendict command fails, it outputs E_WARNING,
+            // Suppress it and handle error internally. assume that it is dictionary open failure
+            if ($dict = @crack_opendict($lcp_crack_dict)) {
+                if (! crack_check($dict,$cp_newpass)) {
+                    $Messages[] = lcp_translate_crack_msg(crack_getlastmessage());
+                    $Messages[] = _("Please choose stronger password.");
+                }
+                crack_closedict($dict);
+            } else {
+                $Messages[] = sprintf(_("Could not open crack dictionary: %s"),$lcp_crack_dict);
+            }
+        }
+    }
+
     if (count($Messages)) {
         return $Messages;
     }
@@ -59,41 +84,13 @@ function change_ldappass_check() {
  * @param string $cp_newpass
  */
 function change_ldappass_go($cp_oldpass, $cp_newpass) {
-    global $username;
-
-    // Set default values
-    $ldap_server = 'localhost';
-    $ldap_protocol_version = 2;
-    $ldap_base_dn = '';
-    $ldap_password_field = 'userpassword';
-    $ldap_user_field = 'uid';
-    $query_dn = '';
-    $query_pw = '';
-    $ldap_filter = '';
-
-    $no_bind_as_user = false;
-
-    $ldap_bind_as_manager = false;
-    $ldap_manager_dn = '';
-    $ldap_manager_pw = '';
-
-    $change_ldapsmb = false;
-    $mkntpwd = '';
-    $ldapsmb_ntpassword = '';
-    $ldapsmb_lmpassword = '';
-
-    $change_smb = false;
-    $smb_host = '';
-    $smb_passwd = '';
-
-    $debug = false;
-    // load config
-    include(SM_PATH . 'plugins/change_ldappass/config.php');
-
-    // field names use lower case
-    $ldap_password_field = strtolower($ldap_password_field);
-    $ldap_user_field = strtolower($ldap_user_field);
-
+    global $username, $ldapsmb_object, $force_smb_sync, $ldap_server,
+        $ldap_protocol_version, $ldap_base_dn, $ldap_password_field,
+        $ldap_user_field, $query_dn, $query_pw, $ldap_filter,
+        $no_bind_as_user, $ldap_bind_as_manager, $ldap_manager_dn,
+        $ldap_manager_pw, $change_ldapsmb, $mkntpwd,
+        $ldapsmb_ntpassword, $ldapsmb_lmpassword, $change_smb,
+        $smb_host, $smb_passwd, $debug;
 
     $Messages = array();
     if ($debug) {
@@ -141,7 +138,10 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
     }
 
     if (!empty($ldap_filter)) {
-        $filter = "(&($ldap_filter)($ldap_user_field=$username))";
+        if (!preg_match('/^\\(.+\\)$/',$ldap_filter)) {
+            $ldap_filter = '(' . $ldap_filter . ')';
+        }
+        $filter = "(&$ldap_filter($ldap_user_field=$username))";
     } else {
         $filter = "($ldap_user_field=$username)";
     }
@@ -177,7 +177,7 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
         array_push($Messages,ldap_debug_print_array($info));
     }
     $dn=$info[0]["dn"]; //finally get the full users dn
-    
+
     // now rebind to the database as user to verify password.
     if (! $no_bind_as_user) {
         if (! @ldap_bind($ds,$dn,$cp_oldpass)) { //if we can't bind as the user then the old passwd must be wrong
@@ -190,7 +190,7 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
             return $Messages;
         } else {
             if ($debug) {
-                array_push($Messages, 
+                array_push($Messages,
                            'LDAP bind successful.',
                            htmlspecialchars("BIND-DN: $dn"));
             }
@@ -204,16 +204,16 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
             }
             return $Messages;
         } else {
-            if ($debug) { 
-                array_push($Messages, 
+            if ($debug) {
+                array_push($Messages,
                            'LDAP bind successful.',
                            htmlspecialchars("BIND-DN: $ldap_manager_dn"));
             }
         }
     }
-    //check the db again, this time so we get the password field returned 
+    //check the db again, this time so we get the password field returned
     // (use ldap_read() in order to lookup only located dn entry)
-    $sr=@ldap_read($ds,$dn,"($ldap_user_field=$username)",array('dn',$ldap_password_field,$ldap_user_field)); 
+    $sr=@ldap_read($ds,$dn,"($ldap_user_field=$username)",array('dn',$ldap_password_field,$ldap_user_field,'objectclass'));
     $info = ldap_get_entries($ds, $sr);
     if ($debug) {
         array_push($Messages,ldap_debug_print_array($info));
@@ -229,19 +229,19 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
         }
         return $Messages;
     }
-    
+
     //this next code tries to identify the correct password type
     //*Note* I've not tested all types here as I cannot reproduce some setups
     //Please let me know if the code does not work for you, or if you have code that will work
     //for a particular type.
-    
+
     //password types:
     //{crypt} salted passwords of type DES, MD5 and Blowfish
     //{MD5} unsalted password in MD5 format
     //{SHA} unsalted password in SHA format
     //{SMD5} salted md5
     //{SSHA} salted sha
-    
+
     //lets try to determine the encrytion method of the stored password
     $p=split("}",$storedpass); //split the password
     $ctype=strtolower($p[0]);  //into the {crypttype}
@@ -251,7 +251,7 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
     case "{crypt":
         $pl=strlen($lpass); // We'll look at the length and salt of what's already stored to determine the crypt sub-type
         $stype="DES";       //sensible default if not detected
-        if ($pl>=34 and substr($lpass,0,3)=="$1$") $stype="MD5"; 
+        if ($pl>=34 and substr($lpass,0,3)=="$1$") $stype="MD5";
         if ($pl>=34 and substr($lpass,0,3)=="$2$") $stype="BLOWFISH";
         if ($debug) array_push($Messages, _("Password type is") . " {crypt}, sub-type $stype");
         $cpass=ldap_crypt_passwd($cp_oldpass,$lpass,$stype); //crypt up our old password so we can check it again
@@ -330,9 +330,19 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
     $newinfo=array();
     $newinfo[$ldap_password_field][0]=$newpass;
 
-    if ($change_ldapsmb) {
+    if (isset($info[0]['objectclass'])) {
+        $objects = $info[0]['objectclass'];
+        array_walk($objects,'lcp_arraytolower');
+    } else {
+        // failsafe for missing objectclass data
+        $objects = array();
+    }
+    if ($change_ldapsmb && in_array($ldapsmb_object,$objects) &&
+        ($force_smb_sync || sqgetGlobalVar('sync_smb_pass',$sync_smb_pass,SQ_POST))) {
         $exe = "$mkntpwd " . escapeshellarg($cp_newpass) . " 2>&1" ;
         if ($debug) array_push($Messages,$exe);
+        $retarray = array();
+        $retval = 1;
         $ntString = exec($exe, $retarray, $retval);
         if ($debug) $Messages = array_merge($Messages,$retarray);
         if ( $retval == "0" && preg_match("/^[0-9A-F]+:[0-9A-F]+$/",$ntString )) {
@@ -340,28 +350,55 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
             $newinfo[$ldapsmb_ntpassword] = $ntPassword;
             $newinfo[$ldapsmb_lmpassword] = $lmPassword;
         } else {  // could not generate ntlm passwords
-            array_push($Messages, _("SMB Password change was not successful, so LDAP not changed!"));
+            array_push($Messages,_("Could not generate NTLM password hashes!"));
+            if (!empty($retarray)) {
+                $retmsg = implode("\n",$retarray);
+                array_push($Messages,sprintf(_("Error: %s"),htmlspecialchars($retmsg)));
+            }
             return $Messages;
         }
     }
 
     if (@ldap_modify($ds,$dn,$newinfo)) {
         $smb=0;
-        if ($change_smb) {
-            if ($smb_host) {
-                $exe="echo " . escapeshellarg($cp_oldpass)  
-                    .  " |$smb_passwd -r " . escapeshellarg($smb_host)
-                    . " -U " . escapeshellarg($username) 
-                    . " -s " . escapeshellarg($cp_newpass) . " 2>&1" ;
+        if ($change_smb &&
+            ($force_smb_sync || sqgetGlobalVar('sync_smb_pass',$sync_smb_pass,SQ_POST))) {
+            // First we print three lines in stdin
+            $exe = 'echo -e ' . escapeshellarg($cp_oldpass . "\\n" . $cp_newpass . "\\n" . $cp_newpass);
+            // then pipe to smbpasswd
+            $exe.= " | $smb_passwd ";
+            // add remote machine name
+            if (!empty($smb_host)) {
+                $exe.= '-r ' . escapeshellarg($smb_host);
+            }
+            // add username, get passwords from stdin (-s), redirect output to stdout
+            $exe.= " -U " . escapeshellarg($username)
+                 . " -s 2>&1";
+
+            // Save used command for debugging
+            if ($debug) {
+                array_push($Messages,$exe);
+            }
+
+            // initial values
+            $retarr = array();
+            $retval = 1;
+            // exec returns last line, but we don't need it. Same line is stored in $retarr
+            exec($exe,$retarr,$retval);
+            // Check $retval. 0 = success, 1 = failure
+            if ($retval) {
+                // push $retarr to $Messages
+                foreach($retarr as $retline) {
+                    array_push($Messages, htmlspecialchars($retline));
+                }
             } else {
-                $exe="echo " . escapeshellarg($cp_oldpass)  
-                    .  " |$smb_passwd -U " . escapeshellarg($username) 
-                    . " -s " . escapeshellarg($cp_newpass) . " 2>&1" ;
+                $smb=1 ;
+                if ($debug) {
+                    foreach($retarr as $retline) {
+                        array_push($Messages, htmlspecialchars($retline));
+                    }
+                }
             }
-            if ($debug) array_push($Messages,$exe);
-            $r=exec($exe,$s);
-            if ($r==_("Password changed for user") . " $username") $smb=1 ;
-            if ($debug) array_push($Messages, $r);
         } else {
             $smb=1;
         }
@@ -379,7 +416,6 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
             if (! $debug) {
                 array_push($Messages, _("Password changed successfully"));
                 return $Messages ;
-                exit(0);
             }
             return $Messages;
         } else { //smbpasswd change failed so we must re sync the ldap password back to its original
@@ -396,10 +432,9 @@ function change_ldappass_go($cp_oldpass, $cp_newpass) {
         if ($debug) array_push($Messages, _("LDAP ERROR => " . ldap_error($ds))) ;
         return $Messages;
     }
-    @ldap_close($ds);
 }
 
-// Generate an unsalted SHA1 pw. This should work withNetscape messaging / directory server 4+ 
+// Generate an unsalted SHA1 pw. This should work withNetscape messaging / directory server 4+
 function ldap_sha_passwd($clear_pw) {
     if(function_exists('sha1')) {
         $hash = pack("H*",sha1($clear_pw));
@@ -446,27 +481,43 @@ function ldap_smd5_passwd($clear_pw) {
     return $new_password ;
 }
 
-// Generate salted crypt passwords
+/**
+ * Generates salted crypt passwords
+ *
+ * @param string $password
+ * @param string $salt
+ * @param string $stype Password type. MD5, BLOWFISH, DES. If other string is
+ * used, since 2.2 function defaults to DES.
+ * @return string
+ */
 function ldap_crypt_passwd($password,$salt,$stype) {
     if ($stype=="MD5")      return crypt($password,substr($salt,0,12)); //MD5 uses 12 chr salt
     if ($stype=="BLOWFISH") return crypt($password,substr($salt,0,16)); //BLOWFISH uses 16 chr salt
-    if ($stype=="DES")  return crypt($password,substr($salt,0,2));      //crypt uses 2 chr salt
+    // DES or something else
+    return crypt($password,substr($salt,0,2));      //crypt uses 2 chr salt
 }
 
+/**
+ * Generates salt
+ * @param string $stype Salt type. MD5, BLOWFISH or DES. If other string is
+ * used, since 2.2 function defaults to DES.
+ * @return string Salt string
+ */
 function ldap_generate_salt($stype) {
     $salt=""; //generate a salt using characters [A-Z][a-z][0-9]./
     $chars="./0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
     for ($i=0;$i<16;$i++) {
-        @ $salt.=$chars[mt_rand(0,strlen($chars))];
+        $salt.=$chars[mt_rand(0,strlen($chars))];
     }
     //return salts longer than we need but these will be trimmed in the ldap_crypt_passwd function
     if ($stype=="MD5")      return "$1$".$salt;
     if ($stype=="BLOWFISH") return "$2$".$salt;
-    if ($stype=="DES")      return $salt;
+    // DES or something else
+    return $salt;
 }
 
 /**
- * This and md5 implimentation of the Salted S2K algorithm as 
+ * This and md5 implimentation of the Salted S2K algorithm as
  * specified in the OpenPGP document (RFC 2440).
  * basically a non mhash dependant version of mhash_keygen_s2k
  * @param string $pass
@@ -509,3 +560,96 @@ function ldap_debug_print_array1($array,$out) {
     }
     return $out;
 }
+
+/**
+ * Translates PHP Crack extension messages
+ *
+ * cpw_translate_crack_msg() function from NaSMail change_password plugin.
+ * @param string $err error from crack_getlastmessage()
+ * @return string translated error message or original error
+ * @since 2.2
+ */
+function lcp_translate_crack_msg($err) {
+    /**
+     * PHP crack 0.4 extension errors
+     * from cracklib_fascist_look_ex()
+     * - it's WAY too short
+     * - it is too short
+     * - it is too simplistic/systematic
+     * - it looks like a National Insurance number.
+     * - it is all whitespace
+     * - it does not contain enough DIFFERENT characters
+     * - it is based on a dictionary word
+     * - it is based on a (reversed) dictionary word
+     * from cracklib_fascist_gecos() (checks are not executed on Win32)
+     * - you are not registered in the password file
+     * - it is based on your username
+     * - it's derivable from your password entry
+     * - it is derivable from your password entry
+     * - it is derived from your password entry
+     * - it's derived from your password entry
+     * - it is based upon your password entry
+     */
+
+    // can't use strcasecmp(). No locale insensitive string comparison functions in PHP.
+    // internal implementation will slow things down.
+    switch($err) {
+    case 'it\'s WAY too short':
+    case 'it is too short':
+        $err = _("Password is too short.");
+        break;
+    case 'it is too simplistic/systematic':
+        $err = _("New password is too simplistic/systematic.");
+        break;
+    case 'it looks like a National Insurance number.':
+        // password looks like personal identification number used
+        // in UK's social security system (aa000000a).
+        $err = _("New password looks like a National Insurance number.");
+        break;
+    case 'it is all whitespace':
+        $err = _("New password contains only whitespace.");
+        break;
+    case 'it does not contain enough DIFFERENT characters':
+        $err = _("New password does not contain enough DIFFERENT characters.");
+        break;
+    case 'it is based on a dictionary word':
+        $err = _("New password is based on a dictionary word.");
+        break;
+    case 'it is based on a (reversed) dictionary word':
+        $err = _("New password is based on a (reversed) dictionary word.");
+        break;
+    case 'you are not registered in the password file':
+        $err = _("You are not registered in the password file.");
+        break;
+    case 'it is based on your username':
+        $err = _("New password is based on your username.");
+        break;
+    case 'it\'s derivable from your password entry':
+    case 'it is derivable from your password entry':
+    case 'it is derived from your password entry':
+    case 'it\'s derived from your password entry':
+    case 'it is based upon your password entry':
+        // combines five error messages.
+        // password is derivable/derived/based upon your password entry
+        $err = _("New password is similar to your current password entry.");
+        break;
+    default:
+        // leave $err untranslated
+    }
+    // 'No obscure checks in this session' - error comes from PHP crack extension
+    // should never pop out because we always call crack_check() before crack_getlastmessage()
+    return $err;
+}
+
+/**
+ * Callback function to lowercase array keys and values
+ *
+ * See ldq_arraytolower in ldapquery plugin.
+ * @param string Array value
+ * @param string Array key
+ * @since 2.2
+ */
+function lcp_arraytolower(&$value,&$key) {
+    $value=strtolower($value);
+    $key=strtolower($key);
+}
